@@ -44,7 +44,6 @@
 #include "htmlengine-edit-selection-updater.h"
 #include "htmlimage.h"
 #include "htmlinterval.h"
-#include "htmllinktext.h"
 #include "htmlobject.h"
 #include "htmlplainpainter.h"
 #include "htmltable.h"
@@ -732,8 +731,9 @@ delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoD
 		    || (e->mark->object->parent && e->mark->object->parent->parent && HTML_IS_TABLE_CELL (e->mark->object->parent->parent))) {
 			check_table_0 (e);
 			check_table_1 (e);
+			html_engine_edit_selection_updater_update_now (e->selection_updater);
 		}
-		if (e->cursor->position == e->mark->position) {
+		if (!html_engine_is_selection_active (e) || e->cursor->position == e->mark->position) {
 			html_engine_disable_selection (e);
 			html_cursor_jump_to_position (e->cursor, e, end_position);
 			return 0;
@@ -769,14 +769,6 @@ delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoD
 	return 0;
 }
 
-void
-html_engine_delete (HTMLEngine *e)
-{
-	html_undo_level_begin (e->undo, "Delete", "Undelete");
-	delete_object (e, NULL, NULL, HTML_UNDO_UNDO, TRUE);
-	html_undo_level_end (e->undo);
-}
-
 gint
 html_engine_cut (HTMLEngine *e)
 {
@@ -883,8 +875,7 @@ insert_object_do (HTMLEngine *e, HTMLObject *obj, guint *len, gint level, guint
 	printf ("position after merge %d\n", e->cursor->position);
 #endif
 
-	html_cursor_destroy (e->cursor);
-	e->cursor = html_cursor_dup (orig);
+	html_cursor_copy (e->cursor, orig);
 	html_cursor_jump_to_position_no_spell (e->cursor, e, position_after);
 
 	if (check)
@@ -1126,7 +1117,8 @@ static void
 check_magic_link (HTMLEngine *e, const gchar *text, guint len)
 {
 	if (HTML_IS_TEXT (e->cursor->object)
-	    && GTK_HTML_PROPERTY (e->widget, magic_links) && len == 1
+	    && gtk_html_get_magic_links (e->widget) 
+	    && len == 1
 	    && (*text == ' ' || text [0] == '\n' || text [0] == '>' || text [0] == ')'))
 		html_text_magic_link (HTML_TEXT (e->cursor->object), e, html_object_get_length (e->cursor->object));
 }
@@ -1151,7 +1143,7 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo)
 	remove_empty_and_merge (e, FALSE, left, right, orig);
 
 	/* replace empty link in empty flow by text with the same style */
-	if (HTML_IS_LINK_TEXT (e->cursor->object) && html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))) {
+	/* FIXME-link if (HTML_IS_LINK_TEXT (e->cursor->object) && html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))) {
 		HTMLObject *flow = e->cursor->object->parent;
 		HTMLObject *new_text;
 
@@ -1166,7 +1158,7 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo)
 			orig->object = e->cursor->object;
 		}
 		html_clue_append (HTML_CLUE (flow), e->cursor->object);
-	}
+		} */
 
 	html_cursor_forward (e->cursor, e);
 
@@ -1206,8 +1198,8 @@ html_engine_insert_empty_paragraph (HTMLEngine *e)
 	insert_empty_paragraph (e, HTML_UNDO_UNDO, TRUE);
 }
 
-static char *picto_chars = "DO)(|/P\0:-\0:\0:-\0:\0:;=-\0:;\0:-~\0:\0:\0:-\0:\0:-\0:\0:-\0:\0";
-static gint picto_states [] = { 8, 13, 18, 26, 34, 39, 44, 0, -1, 11, 0, -1, 0, -2, 16, 0, -2, 0, -3, -4, -5, 23, 0, -3, -4, 0, -6, 30, 32, 0, -6, 0, -11, 0, -8, 37, 0, -8, 0, -9, 42, 0, -9, 0, -10, 47, 0, -10, 0};
+static char *picto_chars = "DO)(|/PQ\0:-\0:\0:-\0:\0:;=-\0:;\0:-~\0:\0:\0:-\0:\0:-\0:\0:-\0:\0:-\0:\0";
+static gint picto_states [] = { 9, 14, 19, 27, 35, 40, 45, 50, 0, -1, 12, 0, -1, 0, -2, 17, 0, -2, 0, -3, -4, -5, 24, 0, -3, -4, 0, -6, 31, 33, 0, -6, 0, -11, 0, -8, 38, 0, -8, 0, -9, 43, 0, -9, 0, -10, 48, 0, -10, 0, -12, 53, 0, -12, 0};
 static gchar *picto_images [] = {
 	"smiley-1.png",
 	"smiley-2.png",
@@ -1220,6 +1212,7 @@ static gchar *picto_images [] = {
 	"smiley-9.png",
 	"smiley-10.png",
 	"smiley-11.png",
+	"smiley-12.png",
 };
 
 static void
@@ -1230,7 +1223,7 @@ use_pictograms (HTMLEngine *e)
 	gint relative;
 	gunichar uc;
 
-	if (!html_object_is_text (e->cursor->object))
+	if (!html_object_is_text (e->cursor->object) || !gtk_html_get_magic_smileys (e->widget))
 		return;
 
 	pos = e->cursor->offset - 1;
@@ -1278,7 +1271,7 @@ use_pictograms (HTMLEngine *e)
 }
 
 void
-html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
+html_engine_insert_text_with_attributes (HTMLEngine *e, const gchar *text, guint len, PangoAttrList *attrs)
 {
 	gchar *nl;
 	gint alen;
@@ -1293,7 +1286,7 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 	gtk_html_editor_event_command (e->widget, GTK_HTML_COMMAND_INSERT_PARAGRAPH, TRUE);
 
 	do {
-		nl   = g_utf8_strchr (text, '\n');
+		nl   = g_utf8_strchr (text, -1, '\n');
 		alen = nl ? g_utf8_pointer_to_offset (text, nl) : len;
 		if (alen) {
 			HTMLObject *o;
@@ -1306,6 +1299,8 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 				html_engine_set_insertion_link (e, NULL, NULL);
 
 			o = html_engine_new_text (e, text, alen);
+			if (attrs)
+				pango_attr_list_splice (HTML_TEXT (o)->attr_list, attrs, 0, HTML_TEXT (o)->text_len);
 			html_text_convert_nbsp (HTML_TEXT (o), TRUE);
 
 			if (alen == 1 && html_is_in_word (html_text_get_char (HTML_TEXT (o), 0))
@@ -1317,7 +1312,7 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 			}
 			insert_object (e, o, html_object_get_length (o), e->cursor->position + html_object_get_length (o),
 				       1, HTML_UNDO_UNDO, check);
-			if (alen == 1 && !HTML_IS_PLAIN_PAINTER (e->painter) && GTK_HTML_PROPERTY (e->widget, magic_smileys))
+			if (alen == 1 && !HTML_IS_PLAIN_PAINTER (e->painter))
 				use_pictograms (e);
 		}
 		if (nl) {
@@ -1330,7 +1325,13 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 }
 
 void
-html_engine_paste_text (HTMLEngine *e, const gchar *text, guint len)
+html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
+{
+	html_engine_insert_text_with_attributes (e, text, len, NULL);
+}
+
+void
+html_engine_paste_text_with_attributes (HTMLEngine *e, const gchar *text, guint len, PangoAttrList *attrs)
 {
 	gchar *undo_name = g_strdup_printf ("Paste text: '%s'", text);
 	gchar *redo_name = g_strdup_printf ("Unpaste text: '%s'", text);
@@ -1339,10 +1340,16 @@ html_engine_paste_text (HTMLEngine *e, const gchar *text, guint len)
 	g_free (undo_name);
 	g_free (redo_name);
 	html_engine_delete (e);
-	html_engine_insert_text (e, text, len);
+	html_engine_insert_text_with_attributes (e, text, len, attrs);
 	html_undo_level_end (e->undo);
 }
 
+void
+html_engine_paste_text (HTMLEngine *e, const gchar *text, guint len)
+{
+	html_engine_paste_text_with_attributes (e, text, len, NULL);
+}
+
 void
 html_engine_delete_container (HTMLEngine *e)
 {
@@ -1636,3 +1643,77 @@ html_engine_cut_and_paste (HTMLEngine *e, const gchar *undo_op_name, const gchar
 		html_object_forall (e->clipboard, e, iterator, data);
 	html_engine_cut_and_paste_end (e);
 }
+
+static void
+delete_upto (HTMLEngine *e, HTMLCursor **start, HTMLCursor **end, HTMLObject *object, guint offset)
+{
+	guint position;
+
+	if (e->mark)
+		html_cursor_destroy (e->mark);
+	e->mark = *start;
+	html_cursor_jump_to (e->cursor, e, object, offset);
+	position = e->cursor->position;
+	delete_object (e, NULL, NULL, HTML_UNDO_UNDO, TRUE);
+	*start = html_cursor_dup (e->cursor);
+	html_cursor_forward (*start, e);
+	(*end)->position -= position - e->cursor->position;
+}
+
+void
+html_engine_delete (HTMLEngine *e)
+{
+	html_undo_level_begin (e->undo, "Delete", "Undelete");
+	html_engine_edit_selection_updater_update_now (e->selection_updater);
+	if (html_engine_is_selection_active (e)) {
+		HTMLCursor *start = html_cursor_dup (e->mark->position < e->cursor->position ? e->mark : e->cursor);
+		HTMLCursor *end = html_cursor_dup (e->mark->position < e->cursor->position ? e->cursor : e->mark);
+		gint start_position = start->position;
+
+		while (start->position < end->position) {
+			if (start->object->parent->parent == end->object->parent->parent) {
+				if (e->mark)
+					html_cursor_destroy (e->mark);
+				html_cursor_destroy (e->cursor);
+				e->mark = start;
+				e->cursor = end;
+				start = end = NULL;
+				delete_object (e, NULL, NULL, HTML_UNDO_UNDO, TRUE);
+				break;
+			} else {
+				HTMLObject *prev = NULL, *cur = start->object;
+
+				/* go thru current cluev */
+				do {
+				        /* go thru current flow */
+					while (cur) {
+				                /* lets look if container is whole contained in the selection */
+						if (html_object_is_container (cur)) {
+							html_cursor_jump_to (e->cursor, e, cur, html_object_get_length (cur));
+							if (e->cursor->position > end->position) {
+								/* it's not => delete upto this container */
+
+								delete_upto (e, &start, &end, cur, 0);
+								prev = NULL;
+								break;
+							}
+						}
+						prev = cur;
+						cur = html_object_next_not_slave (cur);
+					}
+				} while (prev && prev->parent->next && (cur = html_object_head (prev->parent->next)));
+
+				if (prev)
+				        /* cluev end is in the selection */
+					delete_upto (e, &start, &end, prev, html_object_get_length (prev));
+			}
+		}
+
+		if (start)
+			html_cursor_destroy (start);
+		if (end)
+			html_cursor_destroy (end);
+		html_cursor_jump_to_position (e->cursor, e, start_position);
+	}
+	html_undo_level_end (e->undo);
+}
