@@ -2,7 +2,7 @@
 /*  This file is part of the GtkHTML library.
 
     Copyright (C) 2000 Helix Code, Inc.
-    Copyright (C) 2001 Ximian, Inc.
+    Copyright (C) 2001, 2002 Ximian, Inc.
     Authors: Radek Doulik
 
     This library is free software; you can redistribute it and/or
@@ -31,26 +31,19 @@
 #include "htmlengine-edit-cut-and-paste.h"
 #include "htmlengine-edit-table.h"
 #include "htmlengine-edit-tablecell.h"
+#include "htmlinterval.h"
 #include "htmltable.h"
 #include "htmltablecell.h"
 #include "htmltablepriv.h"
 #include "htmlundo.h"
 
-static void delete_table_column (HTMLEngine *e, HTMLUndoDirection dir);
-static void insert_table_column (HTMLEngine *e, gboolean after, HTMLTableCell **column_cells, HTMLUndoDirection dir);
-static void delete_table_row    (HTMLEngine *e, HTMLUndoDirection dir);
-static void insert_table_row    (HTMLEngine *e, gboolean after, HTMLTableCell **row_cells, HTMLUndoDirection dir);
-
 HTMLTable *
 html_engine_get_table (HTMLEngine *e)
 {
-	if (HTML_IS_TABLE (e->cursor->object))
-		return HTML_TABLE (e->cursor->object);
-	else if (!e->cursor->object->parent
-		 || !e->cursor->object->parent->parent
-		 || !e->cursor->object->parent->parent->parent)
-		return NULL;
-	else if (!HTML_IS_TABLE (e->cursor->object->parent->parent->parent))
+	if (!e->cursor->object->parent
+	    || !e->cursor->object->parent->parent
+	    || !e->cursor->object->parent->parent->parent
+	    || !HTML_IS_TABLE (e->cursor->object->parent->parent->parent))
 		return NULL;
 	else
 		return HTML_TABLE (e->cursor->object->parent->parent->parent);
@@ -132,54 +125,133 @@ html_engine_insert_table (HTMLEngine *e, gint cols, gint rows, gint width, gint
  *  Insert Column
  */
 
-static void
-insert_column_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir, guint position_after)
+struct _InsertCellsUndo {
+	HTMLUndoData data;
+
+	gint pos;
+};
+typedef struct _InsertCellsUndo InsertCellsUndo;
+#define INSERT_UNDO(x) ((InsertCellsUndo *) x)
+
+static HTMLUndoData *
+insert_undo_data_new (gint pos)
 {
-	delete_table_column (e, html_undo_direction_reverse (dir));
+	InsertCellsUndo *ud = g_new0 (InsertCellsUndo, 1);
+
+	html_undo_data_init (HTML_UNDO_DATA (ud));
+	ud->pos = pos;
+
+	return HTML_UNDO_DATA (ud);
 }
 
 static void
-insert_column_setup_undo (HTMLEngine *e, guint position_before, HTMLUndoDirection dir)
+insert_column_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir, guint position_after)
+{
+	html_table_delete_column (html_engine_get_table (e), e, INSERT_UNDO (data)->pos, html_undo_direction_reverse (dir));
+}
+
+static void
+insert_column_setup_undo (HTMLEngine *e, gint col, guint position_before, HTMLUndoDirection dir)
 {
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Insert table column", insert_column_undo_action,
-						    NULL, html_cursor_get_position (e->cursor),
+						    insert_undo_data_new (col), html_cursor_get_position (e->cursor),
 						    position_before),
 			      dir);
 }
 
-static void
-insert_table_column (HTMLEngine *e, gboolean after, HTMLTableCell **column, HTMLUndoDirection dir)
+gboolean
+html_engine_goto_table_0 (HTMLEngine *e, HTMLTable *table)
 {
-	HTMLTable *t;
-	HTMLTableCell *cell;
-	gint c, r, col, delta = 0, row, first_row = -1;
-	guint position_before;
+	return html_cursor_jump_to (e->cursor, e, HTML_OBJECT (table), 0);
+}
 
-	t = HTML_TABLE (html_object_nth_parent (e->cursor->object, 3));
-	if (!t)
-		return;
+gboolean
+html_engine_goto_table (HTMLEngine *e, HTMLTable *table, gint row, gint col)
+{
+	HTMLTableCell *cell;
+
+	html_engine_goto_table_0 (e, table);
+	do {
+		cell = html_engine_get_table_cell (e);
+		if (cell && HTML_OBJECT (cell)->parent && HTML_OBJECT (cell)->parent == HTML_OBJECT (table)
+		    && cell->col == col && cell->row == row)
+			return TRUE;
+	} while (cell && html_cursor_forward (e->cursor, e));
+
+	return FALSE;
+}
+
+gboolean
+html_engine_table_goto_col (HTMLEngine *e, HTMLTable *table, gint col)
+{
+	HTMLTableCell *cell;
+
+	if (html_engine_goto_table_0 (e, table)) {
+		cell = html_engine_get_table_cell (e);
+		while (cell && cell->col != col && HTML_OBJECT (cell)->parent == HTML_OBJECT (table)) {
+			html_engine_next_cell (e, FALSE);
+			cell = html_engine_get_table_cell (e);
+		}
+
+		return cell != NULL && HTML_OBJECT (cell)->parent == HTML_OBJECT (table);
+	}
+
+	return FALSE;
+}
+
+gboolean
+html_engine_table_goto_row (HTMLEngine *e, HTMLTable *table, gint row)
+{
+	HTMLTableCell *cell;
+
+	if (html_engine_goto_table_0 (e, table)) {
+		cell = html_engine_get_table_cell (e);
+		while (cell && cell->row != row && HTML_OBJECT (cell)->parent == HTML_OBJECT (table)) {
+			html_engine_next_cell (e, FALSE);
+			cell = html_engine_get_table_cell (e);
+		}
+
+		return cell != NULL && HTML_OBJECT (cell)->parent == HTML_OBJECT (table);
+	}
+
+	return FALSE;
+}
+
+void
+html_table_insert_column (HTMLTable *t, HTMLEngine *e, gint col, HTMLTableCell **column, HTMLUndoDirection dir)
+{
+	HTMLTableCell *cell;
+	HTMLPoint pos;
+	gint c, r;
+	guint position_before;
 
 	html_engine_freeze (e);
 
 	position_before = e->cursor->position;
-	cell = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2)); 
-	col  = cell->col + (after ? 1 : 0);
-	row  = cell->row;
+	pos.object = e->cursor->object;
+	pos.offset = e->cursor->offset;
+	
+	html_engine_goto_table_0 (e, t);
 
 	html_table_alloc_cell (t, 0, t->totalCols);
-
-	for (r = 0; r < t->totalRows; r ++) {
-		for (c = t->totalCols - 1; c > col; c --) {
+	for (c = t->totalCols - 1; c > col; c --) {
+		for (r = 0; r < t->totalRows; r ++) {
 			HTMLTableCell *cell = t->cells [r][c - 1];
 
-			if (cell && cell->col >= col) {
-				if (cell->row == r && cell->col == c - 1)
-					html_table_cell_set_position (cell, r, c);
+			if (cell) {
+				if (cell->col == c - 1) {
+					html_table_cell_set_position (cell, cell->row, c);
+					t->cells [r][c - 1] = NULL;
+				} else if (c == col + 1 && cell->row == r)
+					cell->cspan ++;
+				if (cell->col > c - 1)
+					t->cells [r][c - 1] = NULL;
 				t->cells [r][c] = cell;
-				t->cells [r][c - 1] = NULL;
 			}
 		}
+	}
+	for (r = 0; r < t->totalRows; r ++) {
 		if (!t->cells [r][col]) {
 			guint len;
 
@@ -189,33 +261,12 @@ insert_table_column (HTMLEngine *e, gboolean after, HTMLTableCell **column, HTML
 				: html_engine_new_cell (e, t);
 			html_table_set_cell (t, r, col, cell);
 			html_table_cell_set_position (t->cells [r][col], r, col);
-			if (r < row || (!after && r == row))
-				delta += html_object_get_recursive_length (HTML_OBJECT (cell)) + 1;
-			if (first_row == -1)
-				first_row = r;
 		}
 	}
 
-	e->cursor->position += delta;
-
-	/* now we have right position, let move to the first cell of this new column */
-	if (first_row != -1) {
-		HTMLTableCell *cell;
-		gboolean end;
-		do {
-			cell = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2));
-			if (cell->col == col && cell->row == first_row)
-				break;
-			if (after && first_row >= row)
-				end = html_cursor_forward (e->cursor, e);
-			else
-				end = html_cursor_backward (e->cursor, e);
-		} while (end);
-	} else
-		g_warning ("no new cells added\n");
-
+	html_cursor_jump_to (e->cursor, e, pos.object, pos.offset);
+	insert_column_setup_undo (e, col, position_before, dir);
 	html_object_change_set (HTML_OBJECT (t), HTML_CHANGE_ALL_CALC);
-	insert_column_setup_undo (e, position_before, dir);
 	html_engine_queue_draw (e, HTML_OBJECT (t));
 	html_engine_thaw (e);
 }
@@ -232,7 +283,13 @@ insert_table_column (HTMLEngine *e, gboolean after, HTMLTableCell **column, HTML
 void
 html_engine_insert_table_column (HTMLEngine *e, gboolean after)
 {
-	insert_table_column (e, after, NULL, HTML_UNDO_UNDO);
+	HTMLTable *table;
+	HTMLTableCell *cell;
+
+	table = html_engine_get_table (e);
+	cell = html_engine_get_table_cell (e);
+	if (table && cell)
+		html_table_insert_column (table, e, cell->col + (after ? cell->cspan : 0), NULL, HTML_UNDO_UNDO);
 }
 
 /*
@@ -244,7 +301,7 @@ struct _DeleteCellsUndo {
 
 	HTMLTableCell **cells;
 	gint size;
-	gboolean after;
+	gint pos;
 };
 typedef struct _DeleteCellsUndo DeleteCellsUndo;
 
@@ -259,7 +316,7 @@ delete_cells_undo_destroy (HTMLUndoData *undo_data)
 }
 
 static DeleteCellsUndo *
-delete_cells_undo_new (HTMLTableCell **cells, gint size, gboolean after)
+delete_cells_undo_new (HTMLTableCell **cells, gint size, gint pos)
 {
 	DeleteCellsUndo *data;
 
@@ -269,7 +326,7 @@ delete_cells_undo_new (HTMLTableCell **cells, gint size, gboolean after)
 
 	data->data.destroy = delete_cells_undo_destroy;
 	data->cells        = cells;
-	data->after        = after;
+	data->pos          = pos;
 	data->size         = size;
 
 	return data;
@@ -279,58 +336,58 @@ static void
 delete_column_undo_action (HTMLEngine *e, HTMLUndoData *undo_data, HTMLUndoDirection dir, guint position_after)
 {
 	DeleteCellsUndo *data = (DeleteCellsUndo *) undo_data;
+	HTMLTable *table;
 
-	g_assert (data->size == HTML_TABLE (html_object_nth_parent (e->cursor->object, 3))->totalRows);
-	insert_table_column (e, data->after, data->cells, html_undo_direction_reverse (dir));
+	table = html_engine_get_table (e);
+	g_assert (data->size == table->totalRows);
+	html_table_insert_column (table, e, data->pos, data->cells, html_undo_direction_reverse (dir));
 }
 
 static void
-delete_column_setup_undo (HTMLEngine *e, HTMLTableCell **column, gint size, guint position_before,
-			  gboolean after, HTMLUndoDirection dir)
+delete_column_setup_undo (HTMLEngine *e, HTMLTableCell **column, gint size, guint position_before, gint col, HTMLUndoDirection dir)
 {
-	DeleteCellsUndo *undo_data;
-
-	undo_data = delete_cells_undo_new (column, size, after);
-
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Delete table column", delete_column_undo_action,
-						    HTML_UNDO_DATA (undo_data), html_cursor_get_position (e->cursor),
+						    HTML_UNDO_DATA (delete_cells_undo_new (column, size, col)),
+						    html_cursor_get_position (e->cursor),
 						    position_before), dir);
 }
 
 static void
-delete_table_column (HTMLEngine *e, HTMLUndoDirection dir)
+backward_before_col (HTMLEngine *e, HTMLTable *table, gint col)
+{
+	HTMLObject *cell;
+
+	do {
+		if (!html_cursor_backward (e->cursor, e))
+			return;
+		cell = html_cursor_child_of (e->cursor, HTML_OBJECT (table));
+	} while (cell && HTML_IS_TABLE_CELL (cell) && HTML_TABLE_CELL (cell)->col >= col);
+}
+
+void
+html_table_delete_column (HTMLTable *t, HTMLEngine *e, gint col, HTMLUndoDirection dir)
 {
-	HTMLTable *t;
-	HTMLTableCell *cell;
 	HTMLTableCell **column;
-	HTMLObject *co;
-	gint r, c, col, delta = 0;
+	HTMLTableCell *cell;
+	HTMLPoint pos;
+	gint r, c;
 	guint position_before;
 
-	t = HTML_TABLE (html_object_nth_parent (e->cursor->object, 3));
-
 	/* this command is valid only in table and when this table has > 1 column */
-	if (!t || !HTML_IS_TABLE (HTML_OBJECT (t)) || t->totalCols < 2)
+	if (!t || t->totalCols < 2)
 		return;
 
 	html_engine_freeze (e);
 
 	position_before = e->cursor->position;
-	cell   = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2));
-	col    = cell->col;
 	column = g_new0 (HTMLTableCell *, t->totalRows);
 
-	/* move cursor after/before this column (always keep it in table!) */
-	do {
-		if (col != t->totalCols - 1)
-			html_cursor_forward (e->cursor, e);
-		else
-			html_cursor_backward (e->cursor, e);
-		co = html_object_nth_parent (e->cursor->object, 2);
-	} while (co && co->parent == HTML_OBJECT (t)
-		 && HTML_OBJECT_TYPE (co) == HTML_TYPE_TABLECELL && HTML_TABLE_CELL (co)->col == col);
+	backward_before_col (e, t, col);
+	pos.object = e->cursor->object;
+	pos.offset = e->cursor->offset;
 
+	html_engine_goto_table_0 (e, t);
 	for (r = 0; r < t->totalRows; r ++) {
 		cell = t->cells [r][col];
 
@@ -339,8 +396,6 @@ delete_table_column (HTMLEngine *e, HTMLUndoDirection dir)
 			HTML_OBJECT (cell)->parent = NULL;
 			column [r] = cell;
 			t->cells [r][col] = NULL;
-			delta += html_object_get_recursive_length (HTML_OBJECT (cell));
-			delta ++;
 		}
 
 		for (c = col + 1; c < t->totalCols; c ++) {
@@ -354,12 +409,11 @@ delete_table_column (HTMLEngine *e, HTMLUndoDirection dir)
 		}
 	}
 
-	if (col != t->totalCols - 1)
-		e->cursor->position -= delta;
+	html_cursor_jump_to (e->cursor, e, pos.object, pos.offset);
+	delete_column_setup_undo (e, column, t->totalRows, position_before, col, dir);
 	t->totalCols --;
 
 	html_object_change_set (HTML_OBJECT (t), HTML_CHANGE_ALL_CALC);
-	delete_column_setup_undo (e, column, t->totalRows, position_before, col != t->totalCols - 1, dir);
 	html_engine_queue_draw (e, HTML_OBJECT (t));
 	html_engine_thaw (e);
 }
@@ -374,7 +428,10 @@ delete_table_column (HTMLEngine *e, HTMLUndoDirection dir)
 void
 html_engine_delete_table_column (HTMLEngine *e)
 {
-	delete_table_column (e, HTML_UNDO_UNDO);
+	HTMLTableCell *cell = html_engine_get_table_cell (e);
+
+	if (cell)
+		html_table_delete_column (html_engine_get_table (e), e, cell->col, HTML_UNDO_UNDO);
 }
 
 /*
@@ -384,52 +441,52 @@ html_engine_delete_table_column (HTMLEngine *e)
 static void
 insert_row_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir, guint position_after)
 {
-	delete_table_row (e, html_undo_direction_reverse (dir));
+	html_table_delete_row (html_engine_get_table (e), e, INSERT_UNDO (data)->pos, html_undo_direction_reverse (dir));
 }
 
 static void
-insert_row_setup_undo (HTMLEngine *e, guint position_before, HTMLUndoDirection dir)
+insert_row_setup_undo (HTMLEngine *e, gint row, guint position_before, HTMLUndoDirection dir)
 {
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Insert table row", insert_row_undo_action,
-						    NULL,
+						    insert_undo_data_new (row),
 						    html_cursor_get_position (e->cursor),
 						    html_cursor_get_position (e->cursor)),
 			      dir);
 }
 
-static void
-insert_table_row (HTMLEngine *e, gboolean after, HTMLTableCell **row_cells, HTMLUndoDirection dir)
+void
+html_table_insert_row (HTMLTable *t, HTMLEngine *e, gint row, HTMLTableCell **row_cells, HTMLUndoDirection dir)
 {
-	HTMLTable *t;
 	HTMLTableCell *cell;
-	gint r, c, ntr, row, delta = 0, first_col = -1;
+	HTMLPoint pos;
+	gint r, c;
 	guint position_before;
 
-	t = HTML_TABLE (html_object_nth_parent (e->cursor->object, 3));
-	if (!t)
-		return;
-
 	html_engine_freeze (e);
-
 	position_before = e->cursor->position;
-	cell = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2));
-	row  = cell->row + (after ? 1 : 0);
-	ntr  = t->totalRows + 1;
+	pos.object = e->cursor->object;
+	pos.offset = e->cursor->offset;
+	html_engine_goto_table_0 (e, t);
 
 	html_table_alloc_cell (t, t->totalRows, 0);
-
-	for (c = 0; c < t->totalCols; c ++) {
-		for (r = ntr - 1; r > row; r --) {
+	for (r = t->totalRows; r > row; r --) {
+		for (c = 0; c < t->totalCols; c ++) {
 			HTMLTableCell *cell = t->cells [r - 1][c];
 
-			if (cell && cell->row >= row) {
-				if (cell->row == r - 1 && cell->col == c)
-					html_table_cell_set_position (cell, r, c);
-				t->cells [r][c]     = cell;
-				t->cells [r - 1][c] = NULL;
+			if (cell) {
+				if (cell->row == r - 1) {
+					html_table_cell_set_position (cell, r, cell->col);
+					t->cells [r - 1][c] = NULL;
+				} else if (r == row + 1 && cell->col == c)
+					cell->rspan ++;
+				if (cell->row > r - 1)
+					t->cells [r - 1][c] = NULL;
+				t->cells [r][c] = cell;
 			}
 		}
+	}
+	for (c = 0; c < t->totalCols; c ++) {
 		if (!t->cells [row][c]) {
 			guint len;
 
@@ -439,33 +496,12 @@ insert_table_row (HTMLEngine *e, gboolean after, HTMLTableCell **row_cells, HTML
 				:  html_engine_new_cell (e, t);
 			html_table_set_cell (t, row, c, cell);
 			html_table_cell_set_position (t->cells [row][c], row, c);
-			delta += html_object_get_recursive_length (HTML_OBJECT (cell)) + 1;
-			if (first_col == -1)
-				first_col = c;
 		}
 	}
 
-	if (!after)
-		e->cursor->position += delta;
-
-	/* now we have right position, let move to the first cell of this new row */
-	if (delta) {
-		HTMLTableCell *cell;
-		gboolean end;
-		do {
-			cell = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2));
-			if (cell->col == first_col && cell->row == row)
-				break;
-			if (after)
-				end = html_cursor_forward (e->cursor, e);
-			else
-				end = html_cursor_backward (e->cursor, e);
-		} while (end);
-	} else
-		g_warning ("no new cells added\n");
-
+	html_cursor_jump_to (e->cursor, e, pos.object, pos.offset);
+	insert_row_setup_undo (e, row, position_before, dir);
 	html_object_change_set (HTML_OBJECT (t), HTML_CHANGE_ALL_CALC);
-	insert_row_setup_undo (e, position_before, dir);
 	html_engine_queue_draw (e, HTML_OBJECT (t));
 	html_engine_thaw (e);
 }
@@ -482,8 +518,13 @@ insert_table_row (HTMLEngine *e, gboolean after, HTMLTableCell **row_cells, HTML
 void
 html_engine_insert_table_row (HTMLEngine *e, gboolean after)
 {
-	g_print ("html_engine_insert_table_row\n");
-	insert_table_row (e, after, NULL, HTML_UNDO_UNDO);
+	HTMLTable *table;
+	HTMLTableCell *cell;
+
+	table = html_engine_get_table (e);
+	cell = html_engine_get_table_cell (e);
+	if (table && cell)
+		html_table_insert_row (table, e, cell->row + (after ? cell->rspan : 0), NULL, HTML_UNDO_UNDO);
 }
 
 /*
@@ -494,58 +535,59 @@ static void
 delete_row_undo_action (HTMLEngine *e, HTMLUndoData *undo_data, HTMLUndoDirection dir, guint position_after)
 {
 	DeleteCellsUndo *data = (DeleteCellsUndo *) undo_data;
+	HTMLTable *table;
 
-	g_assert (data->size == HTML_TABLE (html_object_nth_parent (e->cursor->object, 3))->totalCols);
-	insert_table_row (e, data->after, data->cells, html_undo_direction_reverse (dir));
+	table = html_engine_get_table (e);
+	g_assert (data->size == table->totalCols);
+	html_table_insert_row (table, e, data->pos, data->cells, html_undo_direction_reverse (dir));
 }
 
 static void
 delete_row_setup_undo (HTMLEngine *e, HTMLTableCell **row_cells, gint size, guint position_before,
-		       gboolean after, HTMLUndoDirection dir)
+		       gint row, HTMLUndoDirection dir)
 {
-	DeleteCellsUndo *undo_data;
-
-	undo_data = delete_cells_undo_new (row_cells, size, after);
-
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Delete table row", delete_row_undo_action,
-						    HTML_UNDO_DATA (undo_data), html_cursor_get_position (e->cursor),
+						    HTML_UNDO_DATA (delete_cells_undo_new (row_cells, size, row)),
+						    html_cursor_get_position (e->cursor),
 						    position_before), dir);
 }
 
 static void
-delete_table_row (HTMLEngine *e, HTMLUndoDirection dir)
+backward_before_row (HTMLEngine *e, HTMLTable *table, gint row)
+{
+	HTMLObject *cell;
+
+	do {
+		if (!html_cursor_backward (e->cursor, e))
+			return;
+		cell = html_cursor_child_of (e->cursor, HTML_OBJECT (table));
+	} while (cell && HTML_IS_TABLE_CELL (cell) && HTML_TABLE_CELL (cell)->row >= row);
+}
+
+void
+html_table_delete_row (HTMLTable *t, HTMLEngine *e, gint row, HTMLUndoDirection dir)
 {
-	HTMLTable *t;
-	HTMLTableCell *cell;
 	HTMLTableCell **row_cells;
-	HTMLObject *co;
-	gint r, c, row, delta = 0;
+	HTMLTableCell *cell;
+	HTMLPoint pos;
+	gint r, c;
 	guint position_before;
 
-	t = HTML_TABLE (html_object_nth_parent (e->cursor->object, 3));
-
 	/* this command is valid only in table and when this table has > 1 row */
-	if (!t || !HTML_IS_TABLE (HTML_OBJECT (t)) || t->totalRows < 2)
+	if (!t || t->totalRows < 2)
 		return;
 
 	html_engine_freeze (e);
 
 	position_before = e->cursor->position;
-	cell      = HTML_TABLE_CELL (html_object_nth_parent (e->cursor->object, 2));
-	row       = cell->row;
 	row_cells = g_new0 (HTMLTableCell *, t->totalCols);
 
-	/* move cursor after/before this row (always keep it in table!) */
-	do {
-		if (row != t->totalRows - 1)
-			html_cursor_forward (e->cursor, e);
-		else
-			html_cursor_backward (e->cursor, e);
-		co = html_object_nth_parent (e->cursor->object, 2);
-	} while (co && co->parent == HTML_OBJECT (t)
-		 && HTML_OBJECT_TYPE (co) == HTML_TYPE_TABLECELL && HTML_TABLE_CELL (co)->row == row);
+	backward_before_row (e, t, row);
+	pos.object = e->cursor->object;
+	pos.offset = e->cursor->offset;
 
+	html_engine_goto_table_0 (e, t);
 	for (c = 0; c < t->totalCols; c ++) {
 		cell = t->cells [row][c];
 
@@ -554,7 +596,6 @@ delete_table_row (HTMLEngine *e, HTMLUndoDirection dir)
 			HTML_OBJECT (cell)->parent = NULL;
 			row_cells [c] = cell;
 			t->cells [row][c] = NULL;
-			delta += html_object_get_recursive_length (HTML_OBJECT (cell)) + 1;
 		}
 
 		for (r = row + 1; r < t->totalRows; r ++) {
@@ -568,12 +609,10 @@ delete_table_row (HTMLEngine *e, HTMLUndoDirection dir)
 		}
 	}
 
-	if (row != t->totalRows - 1)
-		e->cursor->position -= delta;
+	html_cursor_jump_to (e->cursor, e, pos.object, pos.offset);
 	t->totalRows --;
-
+	delete_row_setup_undo (e, row_cells, t->totalCols, position_before, row, dir);
 	html_object_change_set (HTML_OBJECT (t), HTML_CHANGE_ALL_CALC);
-	delete_row_setup_undo (e, row_cells, t->totalCols, position_before, row != t->totalRows - 1, dir);
 	html_engine_queue_draw (e, HTML_OBJECT (t));
 	html_engine_thaw (e);
 }
@@ -588,7 +627,10 @@ delete_table_row (HTMLEngine *e, HTMLUndoDirection dir)
 void
 html_engine_delete_table_row (HTMLEngine *e)
 {
-	delete_table_row (e, HTML_UNDO_UNDO);
+	HTMLTableCell *cell = html_engine_get_table_cell (e);
+
+	if (cell)
+		html_table_delete_row (html_engine_get_table (e), e, cell->row, HTML_UNDO_UNDO);
 }
 
 typedef enum {
@@ -763,7 +805,7 @@ html_engine_table_set_bg_color (HTMLEngine *e, HTMLTable *t, GdkColor *c)
  *
  */
 
-void table_set_bg_pixmap (HTMLEngine *e, HTMLTable *t, gchar *url, HTMLUndoDirection dir);
+static void table_set_bg_pixmap (HTMLEngine *e, HTMLTable *t, gchar *url, HTMLUndoDirection dir);
 
 static void
 table_set_bg_pixmap_undo_action (HTMLEngine *e, HTMLUndoData *undo_data, HTMLUndoDirection dir, guint position_after)
@@ -773,7 +815,7 @@ table_set_bg_pixmap_undo_action (HTMLEngine *e, HTMLUndoData *undo_data, HTMLUnd
 	table_set_bg_pixmap (e, html_engine_get_table (e), data->attr.pixmap, html_undo_direction_reverse (dir));
 }
 
-void
+static void
 table_set_bg_pixmap (HTMLEngine *e, HTMLTable *t, gchar *url, HTMLUndoDirection dir)
 {
 	HTMLImagePointer *iptr;
@@ -1028,75 +1070,6 @@ html_engine_table_set_width (HTMLEngine *e, HTMLTable *t, gint width, gboolean p
 	table_set_width (e, t, width, percent, HTML_UNDO_UNDO);
 }
 
-gboolean
-html_engine_table_goto_0_0 (HTMLEngine *e)
-{
-	HTMLTableCell *cell;
-
-	cell = html_engine_get_table_cell (e);
-	while (cell && (cell->row != 0 || cell->col != 0)) {
-		html_engine_prev_cell (e);
-		cell = html_engine_get_table_cell (e);
-	}
-
-	return cell != NULL;
-}
-
-gboolean
-html_engine_table_goto_col (HTMLEngine *e, gint col)
-{
-	HTMLTableCell *cell;
-
-	if (html_engine_table_goto_0_0 (e)) {
-		cell = html_engine_get_table_cell (e);
-		while (cell && cell->col != col) {
-			html_engine_next_cell (e, FALSE);
-			cell = html_engine_get_table_cell (e);
-		}
-
-		return cell != NULL;
-	}
-
-	return FALSE;
-}
-
-gboolean
-html_engine_table_goto_row (HTMLEngine *e, gint row)
-{
-	HTMLTableCell *cell;
-
-	if (html_engine_table_goto_0_0 (e)) {
-		cell = html_engine_get_table_cell (e);
-		while (cell && cell->row != row) {
-			html_engine_next_cell (e, FALSE);
-			cell = html_engine_get_table_cell (e);
-		}
-
-		return cell != NULL;
-	}
-
-	return FALSE;
-}
-
-
-gboolean
-html_engine_table_goto_pos (HTMLEngine *e, gint row, gint col)
-{
-	HTMLTableCell *cell;
-
-	if (html_engine_table_goto_0_0 (e)) {
-		cell = html_engine_get_table_cell (e);
-		while (cell && cell->row != row && cell->col != col) {
-			html_engine_next_cell (e, FALSE);
-			cell = html_engine_get_table_cell (e);
-		}
-
-		return cell != NULL;
-	}
-
-	return FALSE;
-}
-
 /*
  * set number of columns in current table
  */
@@ -1113,19 +1086,16 @@ html_engine_table_set_cols (HTMLEngine *e, gint cols)
 		return;
 
 	if (table->totalCols < cols) {
-		html_engine_table_goto_col (e, table->totalCols - 1);
-		while (table->totalCols < cols)
-			html_engine_insert_table_column (e, TRUE);
-	} else {
-		html_engine_table_goto_col (e, table->totalCols - 1);
-		while (table->totalCols > cols)
-			html_engine_delete_table_column (e);
-	}
+		gint n = cols - table->totalCols;
 
-	/*
-	  FIXME: jump to some well defined position (like the same position as before
-	  or at beginning of 1st new column or to last position in remaining cells
-	*/
+		for (; n > 0; n --)
+			html_table_insert_column (table, e, table->totalCols, NULL, HTML_UNDO_UNDO);
+	} else {
+		gint n = table->totalCols - cols;
+
+		for (; n > 0; n --)
+			html_table_delete_column (table, e, table->totalCols - 1, HTML_UNDO_UNDO);
+	}
 }
 
 /*
@@ -1144,18 +1114,16 @@ html_engine_table_set_rows (HTMLEngine *e, gint rows)
 		return;
 
 	if (table->totalRows < rows) {
-		html_engine_table_goto_row (e, table->totalRows - 1);
-		while (table->totalRows < rows)
-			html_engine_insert_table_row (e, TRUE);
-	} else {
-		html_engine_table_goto_row (e, table->totalRows - 1);
-		while (table->totalRows > rows)
-			html_engine_delete_table_row (e);
-	}
+		gint n = rows - table->totalRows;
 
-	/*
-	  FIXME: like set_cols
-	*/
+		for (; n > 0; n --)
+			html_table_insert_row (table, e, table->totalRows, NULL, HTML_UNDO_UNDO);
+	} else {
+		gint n = table->totalRows - rows;
+
+		for (; n > 0; n --)
+			html_table_delete_row (table, e, table->totalRows - 1, HTML_UNDO_UNDO);
+	}
 }
 
 void
