@@ -103,6 +103,7 @@ destroy (HTMLObject *o)
 	g_array_free (table->columnMin, TRUE);
 	g_array_free (table->columnPref, TRUE);
 	g_array_free (table->columnOpt, TRUE);
+	g_array_free (table->columnFixed, TRUE);
 	g_array_free (table->rowHeights, TRUE);
 
 	if (table->bgColor)
@@ -1210,23 +1211,6 @@ draw_background_helper (HTMLTable *table,
 					      paint->y - (o->y - o->ascent));
 }
 
-static void
-draw_background (HTMLObject *self,
-		 HTMLPainter *p,
-		 gint x, gint y, 
-		 gint width, gint height,
-		 gint tx, gint ty)
-{
-	GdkRectangle paint;
-	
-	(* HTML_OBJECT_CLASS (parent_class)->draw_background) (self, p, x, y, width, height, tx, ty);
-
-	if (!html_object_intersect (self, &paint, x, y, width, height))
-	    return;
-
-	draw_background_helper (HTML_TABLE (self), p, &paint, tx, ty);
-}
-
 static void
 draw (HTMLObject *o,
       HTMLPainter *p, 
@@ -1661,6 +1645,8 @@ divide_into_percented_all (HTMLTable *table, gint *col_percent, gint *max_size,
 	for (c = 0; c < table->totalCols; c++)
 		if (active [c] && max_size [c] < ((gdouble) width * PERC (c)) / percent)
 			max_size [c] = ((gdouble) width) * (PERC (c)) / percent;
+	
+	g_free (active);
 }
 
 #define CSPAN (MIN (cell->col + cell->cspan, table->totalCols) - cell->col - 1)
@@ -1992,17 +1978,6 @@ search (HTMLObject *obj, HTMLSearch *info)
 	return FALSE;
 }
 
-static HTMLFitType
-fit_line (HTMLObject *o,
-	  HTMLPainter *painter,
-	  gboolean start_of_line,
-	  gboolean first_run,
-	  gboolean next_to_floating,
-	  gint width_left)
-{
-	return HTML_FIT_COMPLETE;
-}
-
 static void
 append_selection_string (HTMLObject *self,
 			 GString *buffer)
@@ -2286,7 +2261,6 @@ html_table_class_init (HTMLTableClass *klass,
 	object_class->accepts_cursor = accepts_cursor;
 	object_class->calc_size = calc_size;
 	object_class->draw = draw;
-       	object_class->draw_background = draw_background;
 	object_class->destroy = destroy;
 	object_class->calc_min_width = calc_min_width;
 	object_class->calc_preferred_width = calc_preferred_width;
@@ -2298,7 +2272,6 @@ html_table_class_init (HTMLTableClass *klass,
 	object_class->is_container = is_container;
 	object_class->forall = forall;
 	object_class->search = search;
-	object_class->fit_line = fit_line;
 	object_class->append_selection_string = append_selection_string;
 	object_class->head = head;
 	object_class->tail = tail;
